#!/usr/bin/env python -e

"""
docstring
"""

from .operator import *
from .module import *
from .core import *
from .demo import *

__metadata_version__ = "2.1"
__name__ = "volux"
__version__ = "0.9.11"
# __platform__ = "" # (multiple-use) only if the platform is not listed in the “Operating System” Trove classifiers
# __supported_platform__ = #
__summary__ = "High-level media/entertainment workflow automation platform"
__description__ = "High-level media/entertainment workflow automation platform"
__description_content_type__ = "text/markdown"
__keywords__ = "volux media interface workflow automation platform iot lifx volume sound light tk tkinter gui modular"  # A list of additional keywords to be used to assist searching for the distribution in a larger catalog.
__homepage__ = "https://github.com/drtexx/volux"
__author__ = "Denver Pallis"
__author_email__ = "DenverPallisProjects@gmail.com"
# __maintainer__ = "Denver Pallis" # should be omitted if it is identical to Author
# __maintainer_email__ = "DenverPallisProjects@gmail.com" # should be omitted if it is identical to Author
__license__ = "GPLv3+"
__classifiers__ = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Environment :: MacOS X",
    "Environment :: Win32 (MS Windows)",
    "Environment :: X11 Applications",
    "Intended Audience :: Developers",
    "Intended Audience :: End Users/Desktop",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Natural Language :: English",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Topic :: Games/Entertainment",
    "Topic :: Home Automation",
    "Topic :: Multimedia",
    "Topic :: Scientific/Engineering :: Human Machine Interfaces",
    "Topic :: Software Development",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: User Interfaces",
]
__requires_dist__ = [  # requirements
    "lifxlan==1.2.5",
    "pycaw; platform_system == 'Windows'",
    "comtypes==1.1.7; platform_system == 'Windows'",
    "pyalsaaudio==0.8.4; platform_system == 'Linux'",
    "PyAudio==0.2.11; platform_system == 'Linux'",
    "colorama==0.4.1",
    "numpy==1.17.0"
]
__requires_python__ = [">=3"]  # required python version
__requires_external__ = []  # external requirements
__project_urls__ = {  # a browsable URL for the project and a label for it, separated by a comma
    "Bug Reports": "https://github.com/drtexx/{}/issues".format(__name__),
    "Source": "https://github.com/drtexx/{}".format(__name__),
    "Funding": "https://paypal.me/denverpallis",
    "Docs": "https://{}.readthedocs.io".format(__name__),
}

# metadata standard: https://packaging.python.org/specifications/core-metadata/



"""
lt module from lifxtools package
"""

print("lifxtools/lt/__init__.py")

# UI imports
import tkinter as tk  # for UI
from tkinter import ttk  # for not ugly UI

# Backend imports
import lifxtools  # for controlling lights

class LightShowTab(ttk.Frame):
    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        print(self._name,"PARENT =",self.parent)

        self.placeholder = ttk.Label(self, text="PLACEHOLDER")
        self.placeholder.pack()

        self.fade_slider = ttk.Scale(self, from_=0, to=1000)
        self.fade_slider.pack()

        self.minbrightness_slider = ttk.Scale(self, from_=1, to=100)
        self.minbrightness_slider.pack()

        # w = Scale(master, from_=0, to=200, orient=tk.HORIZONTAL)
        # w.pack()


class Navbar(ttk.Frame):
    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        print(self._name,"PARENT =",self.parent)


        self.notebook = ttk.Notebook(self.parent)

        self.page_lightcontrol = ttk.Frame(self.notebook)
        self.page_scenemanager = ttk.Frame(self.notebook)
        self.page_lightshow = LightShowTab(self.notebook)

        self.devices_found = tk.Listbox(self.page_lightcontrol, selectmode=tk.EXTENDED)
        self.devices_found.pack()

        print("devices",self.parent.mlifx.devices) # print found devices to console

        for index, mdevice in enumerate(self.parent.mlifx.managed_devices, 0): # add each found device to the GUI list
            self.devices_found.insert(index, mdevice.label)

        print_selected_device = ttk.Button(self.page_lightcontrol, text="print selected devices index to console", command=self._print_selected_devices)
        print_selected_device.pack()

        toggle_device_power = ttk.Button(self.page_lightcontrol, text="toggle device power", command=self._toggle_device_power)
        toggle_device_power.pack()

        set_to_default = ttk.Button(self.page_lightcontrol, text="set to default", command=self._set_to_default)
        set_to_default.pack()

        set_to_bedtime = ttk.Button(self.page_lightcontrol, text="set to bedtime", command=self._set_to_bedtime)
        set_to_bedtime.pack()

        set_to_theatre = ttk.Button(self.page_lightcontrol, text="set to theatre", command=self._set_to_theatre)
        set_to_theatre.pack()

        save_state = ttk.Button(self.page_lightcontrol, text="save state", command=self._save_state)
        save_state.pack()

        load_state = ttk.Button(self.page_lightcontrol, text="load state", command=self._load_state)
        load_state.pack()



        self.notebook.add(self.page_lightcontrol, text="Light Control")
        self.notebook.add(self.page_scenemanager, text="Scene Manager")
        self.notebook.add(self.page_lightshow, text="Light Show")

        self.notebook.pack(expand=1, fill="both")

    def _print_selected_devices(self):
        print(self.devices_found.curselection()) # return a tuple of indexes for selected items

    def _get_selected_mdevices(self):
        device_indexes = self.devices_found.curselection()
        devices_to_return = [self.parent.mlifx.managed_devices[device_i] for device_i in device_indexes]
        return devices_to_return # return a tuple of indexes for selected items

    def _toggle_device_power(self):
        mdevices = self._get_selected_mdevices()
        print("toggling power of mdevices:",mdevices)
        for mdevice in mdevices:
            mdevice.device.set_power(not mdevice.device.get_power())

    def _set_to_default(self):
        mdevices = self._get_selected_mdevices()
        print("setting mdevices to default:",mdevices)
        for mdevice in mdevices:
            mdevice.device.set_color(lifxtools.default_color)

    def _set_to_bedtime(self):
        mdevices = self._get_selected_mdevices()
        print("setting mdevices to bedtime:",mdevices)
        for mdevice in mdevices:
            mdevice.device.set_color(lifxtools.bedtime_color)

    def _set_to_theatre(self):
        mdevices = self._get_selected_mdevices()
        print("setting mdevices to theatre:",mdevices)
        for mdevice in mdevices:
            mdevice.device.set_color(lifxtools.theatre_color)

    def _save_state(self):
        mdevices = self._get_selected_mdevices()
        print("saving state of mdevices:",mdevices)
        for mdevice in mdevices:
            mdevice.ssave()

    def _load_state(self): # todo: implement the fact you can't load a state that was never saved!
        mdevices = self._get_selected_mdevices()
        print("loading state of mdevices:",mdevices)
        for mdevice in mdevices:
            mdevice.sload()


class DeviceFrameRep(ttk.Frame):
    def __init__(self,mdevice_obj,parent,*args,**kwargs):
        ttk.Frame.__init__(self,parent,*args,**kwargs)
        self.parent = parent

        self.mdevice = mdevice_obj

        self.placeholder = ttk.Label(self,text="DEVICE PREVIEW PLACEHOLDER")
        self.placeholder.pack()

        self.canvas = tk.Canvas(self, width=100, height=100)
        self.canvas.pack()
        self.canvas_rectangle = self.canvas.create_rectangle(0, 0, 100, 100, fill="#40E0D0")

    def refresh(self):
        print(type(self.mdevice))
        print(type(self.mdevice.device))
        color = self.mdevice.device.get_color()
        if (type(color) == tuple):
            h, s, v, k = color
            r, g, b = lifxtools.hsv2rgb(h, s, v)
            hex = "#%02x%02x%02x" % (r, g, b) # todo: fix this please god damn it Denver you fool
            self.canvas.itemconfig(self.canvas_rectangle,fill=hex)
        else:
            print("self.device.color must be a tuple!")


class MainApplication(ttk.Frame):
    def __init__(self, parent, *args, **kwargs):
        ttk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        print(self._name,"PARENT =",self.parent)

        # <create the rest of your GUI here>
        self._init_mlifx()
        self._init_GUI()
        print("finished preparing program!")

    def _init_mlifx(self):

        print("_init_mlifx...")
        self.mlifx = lifxtools.ManagedLifx(lifxtools.return_interface(None))
        self.mlifx.add_device(lifxtools.VirtualDevice())
        print(self.mlifx.devices)

    def _init_GUI(self):

        print("_init_GUI...")
        self.navbar = Navbar(self)
        self.navbar.pack(side="left", fill="y")
        self.device_test = DeviceFrameRep(self.mlifx.managed_devices[0],self) # create a frame for the first device
        self.device_test.pack()

    def _update_loop(self, ms_per_loop=1000):
        # self.device_test.refresh()
        # if live_data == True:
        #     pass
        # self._update_brightness()
        self.after(ms_per_loop, self._update_loop)  # repeat _update_loop()

    def _exit_app(self, event):
        exit()


if __name__ == "lifxtools.lt":

    root = tk.Tk()
    mainApp = MainApplication(root)
    mainApp.pack(side="top", fill="both", expand=True)
    mainApp._update_loop()  # must be before main loop
    root.mainloop()

    # lifx = return_interface(num_lights)
    # devices = lifx.get_lights()
    # list_devices(devices)
    # blink_devices(devices)



class MixerController:

    def __init__(self):

        import os
        self.os = os

    def svol(self,newvol):
        if type(newvol) == int:
            
            if (newvol >= 0) and (newvol <= 100):
                try:
                    self.os.system("osascript -e 'set volume output volume {}'".format(newvol))
                except:
                    raise Exception
            else:
                raise ValueError("volume must be between 0 and 100")
            pass
        else:
            raise TypeError("volume must be an integer")

    def gvol(self):

        return(self.os.system("osascript -e 'set ovol to output volume of (get volume settings)'")) # return average across channels

    def smute(self,newstate):
        
        if type(newstate) == bool:
            
            if newstate == False:
                self.os.system("osascript -e 'set volume with output muted'")
            elif newstate == True:
                self.os.system("osascript -e 'set volume without output muted'")
            else:
                raise Exception("WHAT?")
        
        else:
            raise TypeError("Input must be boolean")

    def gmute(self):
        
        ismuted_string = self.os.system("osascript -e 'output muted of (get volume settings)'")
        
        if ismuted_string == "true":
            return(True)
        elif ismuted_string == "false":
            return(False)
        else:
            raise Exception("ismuted_string didn't return 'true' or 'false', please raise a Github issue! <3")

############################

# set volume on mac
#call(["osascript -e 'set volume output volume {}'".format(volperc)], shell=True)

# get volume on mac
#call(["osascript -e 'set ovol to output volume of (get volume settings)'"], shell=True)



from volux import VoluxModule
import numpy as np
import pyaudio
import colorama
from voluxaudio import lifxtools
from time import sleep
from .hsv2ansi import *

colorama.init()
clamp = lambda value, minv, maxv: max(min(value, maxv), minv)


class VoluxAudio(VoluxModule):
    """
    based on code by the brilliant Scott Harden
    original post code is adapted from:
    https://www.swharden.com/wp/2016-07-19-realtime-audio-visualization-in-python/

    note:
    this code is sensitive to the volume of applications on your desktop.
    Although your actual output audio to speakers shouldn't affect math,
    the volume you're outputting from applications will.
    it is recommended to max out the volume in applications such as spotify
    and then adjust your actual desktop output level to a desirable volume

    protip:
    if you're a spotify user, enable normalization and set your volume level to 'Loud'
    this let's you get the most out of your volume and better maintains audio quality.

    songs for testing:
    Flux Pavilion feat. Steve Aoki - Steve French [looks alright, but beats aren't very solitary]
    Gold Top - Uh Oh (Stand Tall Fists Up Remix) [looks amazing, however the bass is SO INTENSE that multiplier needs to be reduced]
    Yeah Yeah Yeahs - Heads Will Roll (A-Trak Version) (JVH-C Remix) [totally awesome]
    TOP $HELF - Run That $hit [god-tier bass-wobbles and juicy hits]
    Jack U x Ekali x Gravez - Mind (Karol Tip Edit) [insane unique bass that demands attention, inspires me to work on bass isolation so I can replicate the frequency of subs in light saturation]
    SAINT WKND x SAINT MOTEL - MY TYPE [good for testing future bass isolation]
    Jake Hill - Snowflake [my good the audio is the bass in this song, no isolation needed, really clear graph]
    Flume - Burner [thoooose peaaaks are beautiful]
    """

    def __init__(self,sensitivity=9,shared_modules=[],pollrate=None,enable_visualiser=False,*args,**kwargs):
        super().__init__(
            module_name="Volux Audio",
            module_attr="audio",
            module_get=self.get,
            get_type=float,
            get_min=0,
            get_max=100,
            module_set=self.set,
            set_type=float,
            set_min=0,
            set_max=100,
            shared_modules=shared_modules,
            pollrate=pollrate
        )
        self.pa = pyaudio.PyAudio()
        self.stream = self._open_stream()
        self.sensitivity = 9

    def get(self):

        return self._get_amplitude()  # 0 .. 100

    def set(self, new_val):

        print("you can't set the audio amplitude!")

    def _open_stream(self):

        stream = self.pa.open(
            format=pyaudio.paInt16,
            channels=2,  # todo: this shouldn't be hardcoded
            rate=44100,  # todo: this shouldn't be hardcoded
            input=True,
            stream_callback=self._stream_callback,
        )
        return stream

    def _stream_callback(self, in_data, frame_count, time_info, flag):

        audio_data = np.frombuffer(in_data, dtype=np.int16)
        self.audio_data = audio_data
        self.needs_update = True  # is this still needed?

        return (audio_data, pyaudio.paContinue)

    def _get_amplitude(self):
        raw_audio_norm = np.linalg.norm(self.audio_data) / len(self.audio_data)
        adj_audio_norm = raw_audio_norm / 2 ** self.sensitivity
        cla_audio_norm = clamp(adj_audio_norm, 0, 1)  # normalized audio returning a value between 0 to 1
        return cla_audio_norm * 100 # 0 .. 100

    def run_it(self):

        # ---- settings ----

        update_Hz = 240      # 120 is default, low-spec PC's need lower Hz, lights can't
                            # handle Hz too high and crash intermitently
                            # note: tilechains take a lot more processing power to do
                            # realtime music-vis, 60 Hz pretty safe for 1 tilechain
                            # if n in ([n]Hz Δ) is any lower than 0, your computer is
                            # struggling to calculate colors quick enough,
                            # try reducing update_Hz for more accurate music visualisation
                            # note: lifx recommends only sending packets at 50Hz (20ms
                            # per packet) but I'll let you decide if it's worth it
        fade = 20 # in milliseconds
        min_brightness = 1 # default value - 1 (if this is 0 responsiveness might be impacted negatively)
                                   # 65535*0.5 if you want something more bearable on more intense modes
        max_brightness = 65535*1 # default value - 65535
        min_saturation = 65535*0 # default value - 0
        max_saturation = 65535*1 # default value - 65535
        color_temp = 6500 # default: 6500, gnome-NightLight: 4000
        amp_sensitivity = 9 # default: 9 (lower = higher sensitivity, vice versa)
        hue_cycle_duration = 5
        verbose = False

        if (update_Hz < 30):

            print(colorama.Back.RED + colorama.Fore.WHITE + "update_Hz is < 30, enabling automatic fade calculation" + colorama.Style.RESET_ALL)
            sleep(1)

            if (update_Hz >= 1):

                fade = 1000/update_Hz # automatically determine transition duration based on the time between color packets sent to the bulbs (especially makes slower Hz look better)

            else:

                raise ValueError("update_Hz must be greater than or equal to 1!")

        # ---- script ----

        mlifx = lifxtools.ManagedLifx(lifxtools.return_interface(None),verbose=verbose)

        self.lal = lifxtools.LightShow(mlifx,pyaudio,update_Hz,fade,min_brightness=min_brightness,max_brightness=max_brightness,min_saturation=min_saturation,max_saturation=max_saturation,color_temp=color_temp,hue_cycle_duration=hue_cycle_duration,amp_sensitivity=amp_sensitivity)

        self.lal.start()

    def stop(self):

        # self.mlifx.restore()  # restore original state of all managed lights
        self.stream.stop_stream()
        self.stream.close()
        self.pa.terminate()
        print(Style.RESET_ALL)  # reset ansi escapes



from volux import VoluxModule
import tkinter as tk
from tkinter import ttk

class VoluxGui(VoluxModule):
    def __init__(self, shared_modules=[], pollrate=None, *args, **kwargs):
        super().__init__(
            module_name="Volux GUI",
            module_attr="gui",
            module_get=self.get,
            get_type=None,
            get_min=None,
            get_max=None,
            module_set=self.set,
            set_type=None,
            set_min=None,
            set_max=None,
            shared_modules=shared_modules,
            pollrate=pollrate
        )

        self.root = tk.Tk()
        self.mainApp = MainApplication(self.root,self,style="mainApp.TFrame")
        self.example_val = 0

    def get(self):

        return self.example_val

    def set(self, new_val):

        self.example_val = new_val

    def init_window(self):

        self.gui_style = ttk.Style()
        self.gui_style.configure('mainApp.TFrame', background="#19191B")

        self.mainApp.pack(side="top", fill=tk.BOTH, expand=True)
        self.root.title("Volux")
        screen_size = (self.root.winfo_screenwidth(), self.root.winfo_screenheight())
        self.root.geometry("{}x{}+{}+{}".format(
            800,
            500,
            1920-800,
            1080-500
        )) # define the size of the window
        self.mainApp._update_loop()
        self.root.mainloop()

class MainApplication(ttk.Frame):
    def __init__(self,parent,module_root,*args,**kwargs):
        ttk.Frame.__init__(self,parent,*args,**kwargs)
        self.parent = parent
        self.module_root = module_root
        self.ext_modules = self.module_root._shared_modules
        self.vis_on = tk.BooleanVar()

        print("(VoluxGUI) shared modules:",self.ext_modules)

        self.wip_notice = ttk.Label(self,text="NOTE: GUI IS A WORK-IN-PROGRESS",anchor=tk.CENTER)
        self.wip_notice.pack(side="top",fill=tk.X)

        self.input_frame = ttk.Frame(self)
        self.input_frame.pack(side="left",fill=tk.Y,padx="14px",pady="14px")

        self.input_label = ttk.Label(self.input_frame,text="INPUTS")
        self.input_label.pack(side="top",fill=tk.BOTH)

        self.input_listbox = tk.Listbox(self.input_frame, selectmode=tk.SINGLE, exportselection=False)
        self.input_listbox.pack()

        self.input_testget = ttk.Button(self.input_frame,text="Get Test",command=self._get_test)
        self.input_testget.pack()

        self.output_frame = ttk.Frame(self)
        self.output_frame.pack(side="left",fill=tk.Y,padx="14px",pady="14px")

        self.output_label = ttk.Label(self.output_frame,text="OUTPUTS")
        self.output_label.pack(side="top",fill=tk.BOTH)

        self.output_listbox = tk.Listbox(self.output_frame, selectmode=tk.SINGLE, exportselection=False)
        self.output_listbox.pack()

        self.output_testset = ttk.Button(self.output_frame,text="Set Test",command=self._set_test)
        self.output_testset.pack()

        self.output_testset_data = ttk.Entry(self.output_frame)
        self.output_testset_data.pack()

        self.demo_slider = ttk.Scale(self.output_frame, from_=0, to=100)
        self.demo_slider.pack()

        for smodule in self.ext_modules:
            if hasattr(smodule,'tk_frame'):
                setattr(self,smodule.frame_attr,smodule.tk_frame(self))
                this_frame = getattr(self,smodule.frame_attr)
                smodule._set_gui_instance(self.module_root)
                this_frame.pack(side="left",fill=tk.Y,padx="14px",pady="14px")

        self._update_input_listbox()
        self._update_output_listbox()

    def _update_loop(self):

        pass

    def _get_test(self):

        sel = self.input_listbox.curselection()
        if len(sel) > 0:
            i = sel[0]
            get_result = self.ext_modules[i].get()
            print("(VoluxGui) get method response:",get_result)
        else:
            raise ValueError("no input selected!")

    def _set_test(self):

        sel = self.output_listbox.curselection()
        if len(sel) > 0:
            i = self.output_listbox.curselection()[0]
            get_result = self.ext_modules[i].set(int(self.output_testset_data.get()))
            print("(VoluxGui) set method response:",get_result)
        else:
            raise ValueError("no output selected!")

    def _update_input_listbox(self):

        for i in range(len(self.ext_modules)):

            module = self.ext_modules[i]

            if hasattr(module,'get'):

                self.input_listbox.insert(tk.END,module._module_name)

    def _update_output_listbox(self):

        for i in range(len(self.ext_modules)):

            module = self.ext_modules[i]

            if hasattr(module,'set'):

                self.output_listbox.insert(tk.END,module._module_name)

    def _get_selected_mdevices(self):
        devices_to_return = [self.parent.mlifx.managed_devices[device_i] for device_i in device_indexes]
        return devices_to_return # return a tuple of indexes for selected items



#!/usr/bin/env python

from gi.repository import Gtk
import subprocess

class BrightnessScale:
    def __init__(self):
        # get active monitor and current brightness
        self.monitor = self.getActiveMonitor()
        self.currB = self.getCurrentBrightness()

    def initUI(self):
        # initliaze and configure window 
        window = Gtk.Window()
        window.set_title('Brightness Scale')
        window.set_default_size(250, 50)
        window.set_position(Gtk.WindowPosition.CENTER)
        window.set_border_width(10)

        # slider configuration
        self.adjustment = Gtk.Adjustment(self.currB, 0, 100, 1, 10, 0)
        self.scale = Gtk.HScale()
        self.scale.set_adjustment(self.adjustment)
        self.scale.set_digits(0)

        # close Gtk thread on closing window
        window.connect("destroy", lambda w: Gtk.main_quit())

        # setup event handler on value-changed
        self.scale.connect("value-changed", self.scale_moved)

        # add the scale to window
        window.add(self.scale)

        # show all components in window
        window.show_all()

        # close window on pressing escape key
        accGroup = Gtk.AccelGroup()
        key, modifier = Gtk.accelerator_parse('Escape')
        accGroup.connect(key, modifier, Gtk.AccelFlags.VISIBLE, Gtk.main_quit)
        window.add_accel_group(accGroup)

    def showErrDialog(self):
        self.errDialog = Gtk.MessageDialog(None, 
                                           Gtk.DialogFlags.MODAL,
                                           Gtk.MessageType.ERROR,
                                           Gtk.ButtonsType.OK,
                                           "Unable to detect active monitor, run 'xrandr --verbose' on command-line for more info")
        self.errDialog.set_title("brightness control error")
        self.errDialog.run()
        self.errDialog.destroy()

    def initStatus(self):
        if(self.monitor == "" or self.currB == ""):
            return False
        return True

    def getActiveMonitor(self):
        #Find display monitor
        monitor = subprocess.check_output("xrandr -q | grep ' connected' | cut -d ' ' -f1", shell=True)
        print(monitor)
        if(monitor != ""):
            monitor = monitor.split('\n')[0]
        return monitor

    def getCurrentBrightness(self):
        #Find current brightness
        currB = subprocess.check_output("xrandr --verbose | grep -i brightness | cut -f2 -d ' '", shell=True)
        if(currB != ""):
            currB = currB.split('\n')[0]
            currB = int(float(currB) * 100)
        else:
            currB = ""
        return currB

    def scale_moved(self, event):
        #Change brightness
        newBrightness = float(self.scale.get_value())/100
        cmd = "xrandr --output %s --brightness %.2f" % (self.monitor, newBrightness)
        cmdStatus = subprocess.check_output(cmd, shell=True)

if __name__ == "__main__":
    # new instance of BrightnessScale
    brcontrol = BrightnessScale()
    if(brcontrol.initStatus()):
        # if everything ok, invoke UI and start Gtk thread loop
        brcontrol.initUI()
        Gtk.main()
    else:
        # show error dialog
        brcontrol.showErrDialog()


class VoluxModule:
    """Provides common metadata for every Volux module. All modules (including core) are a subclass of VoluxModule"""

    def __init__(self, module_name, module_attr, module_get, get_type,
    get_min, get_max, module_set, set_type, set_min, set_max,
    shared_modules, pollrate):

        if not type(shared_modules) == list:
            raise TypeError("VoluxModule: kwarg 'shared_modules' must be of type list")

        self._module_name = module_name
        self._module_attr = module_attr
        self.get = module_get
        self._get_type = get_type
        self._get_min = get_min
        self._get_max = get_max
        self.set = module_set
        self._set_type = set_type
        self._set_min = set_min
        self._set_max = set_max
        self._shared_modules = shared_modules
        self._pollrate = pollrate

    def _loaded(self):
        print("loaded module: {} (pollrate={})".format(self._module_name,self._pollrate))

    def get_module_info(self):
        return {
            'name': self._module_name,
            'attr': self._module_attr
        }



import warnings
warnings.warn("WARNING: MODULE IS A WORK IN PROGRESS!!!")

class Display:

    def __init__(self):

        import subprocess
        self.subprocess = subprocess
        self.Popen = subprocess.Popen
        self.PIPE = subprocess.PIPE

        self.name = None
        self.brightness = None

    def _gDisplayName(self):

        display_name = self.subprocess.call(['xrandr | grep " connected"'], shell=True)
        print(display_name)
#         process = self.Popen(['xrandr | grep "connected"'], stdout=self.PIPE, stderr=self.PIPE, shell=True)
#         stdout, stderr = process.communicate()
#         print(stdout)

    def sbrightness(self,displayname,new_brightness):

        "xrandr --output {} --brightness {}".format(displayname,new_brightness) # NOTE: CONTROLS LITERAL BRIGHTNESS, NOT BACKLIGHT BRIGHTNESS!!!

    def gbrightness(self):

        pass
        # get the brightness
        # set the self.brightness value


disp = Display()
disp_name = disp._gDisplayName()
disp.sbrightness(disp_name,0.5)



from .core import VoluxCore
from .module import VoluxModule


class VoluxOperator:
    def __init__(self):
        self.modules = []
        self.add_module(VoluxCore())

    def add_module(self, module):

        if self.validate_module(module):  # if the object passed is a valid module

            if not module in self.modules:

                self.modules.append(module)
                setattr(self, module._module_attr, module)
                module._loaded()  # call module's method for when it's finished being loaded

            else:

                raise Exception("module '{}' is already loaded!".format(module._module_name))

        else:

            raise TypeError("module must be a subclass of VoluxModule")

    def remove_module(self, module):

        if module in self.modules:

            self.modules.remove(module)
            delattr(self, module._module_attr)

        else:

            raise AttributeError("module '{}' not loaded!".format(module._module_name))

    def validate_module(self, module):

        for attrib in ["_module_name", "_module_attr", 'get', 'set']:

            if not hasattr(module, attrib):

                print(
                    "module must be a subclass of VoluxModule with the [{}] attribute".format(
                        attrib
                    )
                )
                return False

        return True

    def get_modules(self):

        return self.modules



class MixerController:
    def __init__(self):

        from ctypes import cast, POINTER
        from comtypes import CLSCTX_ALL
        from pycaw.pycaw import AudioUtilities, IAudioEndpointVolume

        devices = AudioUtilities.GetSpeakers()
        interface = devices.Activate(IAudioEndpointVolume._iid_, CLSCTX_ALL, None)

        self.volume = cast(interface, POINTER(IAudioEndpointVolume))
        _vol_range = self.volume.GetVolumeRange()
        self._vol_min = _vol_range[0]
        self._increment = abs(_vol_range[0])/100

    def svol(self,newvol):

        if type(newvol) == int:

            if (newvol >= 0) and (newvol <= 100):

                try:
                    _steps = 100 - newvol
                    newmaster = 0 -(self._increment*_steps)
                    self.volume.SetMasterVolumeLevel(newmaster, None)

                except:
                    raise Exception

            else:
                raise ValueError("volume must be between 0 and 100")
        else:
            raise TypeError("volume must be an integer")

    def gvol(self):

        # get abs min
        _abs_min = abs(self._vol_min)
        # get abs current value
        _abs_val = self.volume.GetMasterVolumeLevel()
        # divide current by abs min
        inverse_percent = _abs_val / _abs_min
        # multiply product by 100 for scaled percentage
        _gvol = 100-(abs(inverse_percent)*100)
        return(round(_gvol))

    def smute(self,newstate):

        if type(newstate) == bool:
             
            if newstate == False:
                self.volume.SetMute(0, None)
            elif newstate == True:
                self.volume.SetMute(1, None)
            else:
                raise Exception("WHAT?")

        else:
            raise TypeError("Input must be boolean")

    def gmute(self):

        _gmute = self.volume.GetMute()
         
        if _gmute == 0:
            return(False)
        elif _gmute == 1:
            return(True)
         
        else:
            raise Exception("Mute didn't return 0 or 1, please create an issue on Github!")



from volux import VoluxModule
import tkinter as tk
from tkinter import ttk

class VoluxBar(VoluxModule):
    def __init__(self,shared_modules=[],pollrate=100,*args,**kwargs):
        super().__init__(
            module_name="Volux Bar",
            module_attr="bar",
            module_get=self.get,
            get_type=int,
            get_min=0,
            get_max=100,
            module_set=self.set,
            set_type=int,
            set_min=0,
            set_max=100,
            shared_modules=shared_modules,
            pollrate=pollrate
        )
        self.modes = {}
        self.mode = 'default'
        self.root = tk.Tk()
        self.mainApp = MainApplication(self.root,self,style='Bar.TFrame')

    def add_mode(self,name,module):
        self.modes.update({name: module})

    def remove_mode(self,name):
        self.modes.pop(name)

    def get(self):
        print("WIP!")

    def set(self, new_val):
        print("WIP!")

    def init_window(self):
        self.gui_style = ttk.Style()
        self.gui_style.configure('BarVal.TFrame', background="GREEN")
        self.gui_style.configure('Bar.TFrame', background="BLACK")

        self.mainApp.pack(side="top", fill=tk.BOTH)
        self.root.title("title")
        self.root.attributes("-topmost",True) # force window to stay on top (doesn't work in full screen applications)
        self.root.overrideredirect(1) # remove frame of window
        screen_size = (self.root.winfo_screenwidth(), self.root.winfo_screenheight())
        self.root.geometry("{}x{}+{}+{}".format(
            1920,
            8,
            0,
            1080-8
        )) # define the size of the window
        self.root.wait_visibility(self.root) # required for window transparency
        self.root.wm_attributes("-alpha",0.1) # make window transparent
        self.mainApp._update_loop()
        self.root.mainloop()


class MainApplication(ttk.Frame):
    def __init__(self,parent,VoluxBar_obj,*args,**kwargs):
        ttk.Frame.__init__(self,parent,*args,**kwargs)
        self.parent = parent
        self.VoluxBar_obj = VoluxBar_obj

        self.s_width = self.parent.winfo_screenwidth()

        self._bar = VolumeBar(self, style='Bar.TFrame')
        self._bar.pack(fill=tk.Y,side=tk.LEFT)

        self._init_bindings()

    def _update_loop(self):
        self.parent.lift() # ensure window on top of others
        self._refresh_bar()
        # print("refreshing...")
        self.after(self.VoluxBar_obj._pollrate,self._update_loop) # repeat _update_loop()

    def _refresh_bar(self):
        val = self.VoluxBar_obj.modes[self.VoluxBar_obj.mode].get()
        self._bar.barValue.config(width=self.s_width*(val/100))
        pass

    def _init_bindings(self):
        self.bind("<Enter>",self._mouse_entered)
        self.bind("<Leave>",self._mouse_left)
        self.parent.bind("<MouseWheel>",self._scroll_delta)
        self.parent.bind("<Button-2>",self._middle_click)
        self.parent.bind("<Button-4>",self._scroll_up)
        self.parent.bind("<Button-5>",self._scroll_down)
        self.parent.bind("<Button-3>",self._right_click)
        # self.parent.bind("<Control-Button-4>",self._brightness_up)
        # self.parent.bind("<Control-Button-5>",self._brightness_down)
        self.parent.bind("<Double-Button-3>",self._exit_app)

    def _mouse_entered(self, event):

        self.parent.wm_attributes("-alpha",0.8) # make window transparent

    def _mouse_left(self, event):

        self.parent.wm_attributes("-alpha",0.1) # make window transparent

    def _middle_click(self, event):

        active_module = self.VoluxBar_obj.modes[self.VoluxBar_obj.mode]

        if hasattr(active_module,'toggle'):
            new_state = active_module.toggle()
            if new_state == True:
                self.VoluxBar_obj.gui_style.configure('BarVal.TFrame', background="RED")
            elif new_state == False:
                self.VoluxBar_obj.gui_style.configure('BarVal.TFrame', background="GREEN")
        else:
            print("module '{}' has no toggle method!".format(active_module._module_name))

    def _scroll_delta(self, event):

        movement = event.delta/120

        if movement == 1:
            self._scroll_up(event)

        elif movement == -1:
            self._scroll_down(event)

        else:
            raise ValueError("movement should be 1 or -1")

    def _scroll_up(self, event):

        val = self.VoluxBar_obj.modes[self.VoluxBar_obj.mode].get()
        self.VoluxBar_obj.modes[self.VoluxBar_obj.mode].set(val+5)
        self._refresh_bar()

    def _scroll_down(self, event):

        val = self.VoluxBar_obj.modes[self.VoluxBar_obj.mode].get()
        self.VoluxBar_obj.modes[self.VoluxBar_obj.mode].set(val-5)
        self._refresh_bar()

    def _right_click(self, event):

        if self.VoluxBar_obj.mode == 'default':

            self.VoluxBar_obj.mode = 'light'
            self.VoluxBar_obj.gui_style.configure('BarVal.TFrame', background="BLUE")

        elif self.VoluxBar_obj.mode == 'light':

            self.VoluxBar_obj.mode = 'default'
            self.VoluxBar_obj.gui_style.configure('BarVal.TFrame', background="GREEN")

        else:

            raise Exception("mode not recognized!")

        self._refresh_bar()

    def _exit_app(self, event):

        exit()


class VolumeBar(ttk.Frame):
    def __init__(self,parent,*args,**kwargs):
        ttk.Frame.__init__(self,parent,*args,**kwargs)
        self.parent = parent

        self.barValue = ttk.Frame(self,width=1920*0.5,height=100,style="BarVal.TFrame")
        self.barValue.pack()



class VirtualDevice:
    def __init__(self, label="Virtual Device", power=True, color=(0,0,65535,6500), is_light=True, supports_color=True):
        self._label = label
        self._power = power
        self._color = color
        self._is_light = is_light
        self._supports_color = supports_color

    def get_label(self):
        return self._label

    def get_power(self):
        return self._power

    def is_light(self):
        return self._is_light

    def supports_color(self):
        return self._supports_color

    def set_power(self,new_power_state):
        self._power = new_power_state
        print("new power state for [{}]: {}".format(self.get_label(),self._power))

    def set_color(self,HSVK_tuple):
        self._color = HSVK_tuple
        print("new color for [{}]: {}".format(self.get_label(),self._color))

    def get_color(self):
        return self._color



from volux import VoluxModule
import cpaudio


class VoluxVolume(VoluxModule):
    def __init__(self,shared_modules=[],pollrate=100,*args,**kwargs):
        super().__init__(
            module_name="Volux Volume",
            module_attr="volume",
            module_get=self.get,
            get_type=int,
            get_min=0,
            get_max=100,
            module_set=self.set,
            set_type=int,
            set_min=0,
            set_max=100,
            shared_modules=shared_modules,
            pollrate=pollrate
        )
        self.mixer = cpaudio.MixerController()

    def get(self):

        return self.mixer.gvol()

    def set(self,new_val):

        if new_val < 0:

            new_val = 0

        elif new_val > 100:

            new_val = 100

        self.mixer.svol(new_val)

    def toggle(self):

        new_state = not self.mixer.gmute()
        self.mixer.smute(new_state)
        return new_state



#!/usr/bin/env python3 -e

import argparse
import colorama
colorama.init()

def main():

    list_choices = ['demos','modules','scripts']
    wip_string = "{}not yet implemented{}".format(colorama.Fore.YELLOW,colorama.Style.RESET_ALL)

    # create the top-level parser
    parser = argparse.ArgumentParser()

    # optional arguments before sub-commands
    parser.add_argument('-V','--version', action="store_true", help="print version information and exit")
    parser.add_argument('--lists', action="store_true", help="show available lists")

    # create subparsers object
    subparsers = parser.add_subparsers(dest="subcommand",help='available sub-commands')

    # create the parser for the "list" command
    parser_list = subparsers.add_parser('list', help='list available demos, modules or scripts')
    parser_list.add_argument('LIST', action="store", help='name of list to retrieve')

    # create the parser for the "demo" command
    parser_demo = subparsers.add_parser('demo', help='run the specified demo')
    parser_demo.add_argument('DEMO', action="store", help='name of the demo to run')
    parser_demo.add_argument('--baz', choices='XYZ', help='baz help')

    # create the parser for the "script" command
    parser_script = subparsers.add_parser('script', help="run the specified script")
    parser_script.add_argument("SCRIPT", action="store", help="name of script to run")

    # create the parser for the "info" command
    parser_info = subparsers.add_parser('info', help="show information for a specified module")
    parser_info.add_argument("MODULE", action="store", help="name of the module to get info for")

    # create the parser for the "launch" command
    parser_launch = subparsers.add_parser('launch', help="launch the volux GUI", description="launches the volux GUI")
    parser_launch.add_argument("--foo", action="store_true", help="a test optional argument")

    # Parse the arguments
    arguments = parser.parse_args()

    if arguments.version == True:

        from volux import __name__, __version__
        print(__name__,__version__)
        exit()

    elif arguments.lists == True:

        print("Available lists:",", ".join(list_choices))
        exit()

    if arguments.subcommand == 'launch':

        print("launching...")
        from volux.scripts import launch_gui
        launch_gui()
        exit()

    elif arguments.subcommand == 'list':

        if arguments.LIST == 'demos':
            from volux import VoluxCore
            _core = VoluxCore()
            demos_string = ",".join(_core.get_demo_aliases())
            print("Available demos:",demos_string)

        elif arguments.LIST == 'modules':
            print("Available modules:",wip_string)

        elif arguments.LIST == 'scripts':
            print("Available scripts:",wip_string)

        else:

            print("{}Error: '{}' is not a valid list name{}".format(colorama.Fore.RED,arguments.LIST,colorama.Style.RESET_ALL))
            print("{}Tip: See 'volux --lists'{}".format(colorama.Fore.YELLOW,colorama.Style.RESET_ALL))

        exit()

    elif arguments.subcommand == 'demo':

        from volux import demos, VoluxDemo, VoluxCore

        _core = VoluxCore()
        demo_dict = _core.get_demo_dict()
        if arguments.DEMO in demo_dict:
            demo_dict[arguments.DEMO].run_demo()
        else:
            print("{}Error: '{}' is not a valid demo{}".format(colorama.Fore.RED,arguments.DEMO,colorama.Style.RESET_ALL))
            print("{}Tip: See 'volux list demos'{}".format(colorama.Fore.YELLOW,colorama.Style.RESET_ALL))

        exit()

    elif arguments.subcommand == 'script':

        print("script launching is {}".format(wip_string))

        exit()

    elif arguments.subcommand == 'info':

        print("module info is {}".format(wip_string))

        exit()

    else:

        print("{}Error: No command specified{}".format(colorama.Fore.RED,colorama.Style.RESET_ALL))
        print("{}Tip: See 'volux --help'{}".format(colorama.Fore.YELLOW,colorama.Style.RESET_ALL))

        exit()

if __name__ == '__main__':

    main()



from .module import VoluxModule

class VoluxCore(VoluxModule):
    """provides a set of utilities for use in other modules"""
    def __init__(self,*args,**kwargs):
        super().__init__(
            module_name="Volux Core",
            module_attr="core",
            module_get=None,
            get_type=None,
            get_min=None,
            get_max=None,
            module_set=None,
            set_type=None,
            set_min=None,
            set_max=None,
            shared_modules=[],
            pollrate=None
        )

    def get_python_module_items(self,module):

        return [getattr(module,item_name) for item_name in dir(module)]

    def filter_by_attr_value(self,items,attribute,attribute_value):
        """returns only items which have [item].[attribute] == [attribute_value]"""

        valid_items = []

        for item in items:

            if hasattr(item,attribute):

                if getattr(item,attribute) == attribute_value:

                    valid_items.append(item)

        return(valid_items)

    def filter_by_superclass(self,items,superclass):
        """return all objects which have inherited a particular superclass"""

        return self.filter_by_attr_value(items, 'superclass', superclass)

    def gen_demo_dict(self,demo_list):
        """turn a list for VoluxDemos into a dictionary of {[demo alias]: [demo]} pairs"""

        demo_dict = {demo._alias: demo for demo in demo_list}
        return demo_dict

    def get_demos(self):

        from volux import demos, VoluxDemo, VoluxCore
        items = self.get_python_module_items(demos) # for each item in the demos module
        demos_collected = self.filter_by_superclass(items,VoluxDemo) # filter out items not inherited from VoluxDemo class
        return demos_collected

    def get_demo_aliases(self):

        demos_collected = self.get_demos()
        demo_aliases = [demo._alias for demo in demos_collected]
        return demo_aliases

    def get_demo_dict(self):

        return {demo._alias: demo for demo in self.get_demos()}



from ..demo import VoluxDemo

class DemoVolLuxBar(VoluxDemo):
    def __init__(self,*args,**kwargs):
        super().__init__(
            demo_name='Demo Vol Lux Bar',
            demo_method=self.run_demo,
            alias="bar",
            *args,
            **kwargs
        )
        self.superclass = VoluxDemo

    def run_demo(self):

        import volux
        from voluxbar import VoluxBar
        from voluxvolume import VoluxVolume
        from voluxlight import VoluxLight

        # create Volux Operator object (hub for communication between modules)
        vlx = volux.VoluxOperator()

        # load Volux Bar module
        vlx.add_module(VoluxBar())
        vlx.add_module(VoluxVolume())
        vlx.add_module(VoluxLight("all_devices", init_mode_args={'label': 'Strip'}, group="Office"))
        vlx.bar.add_mode("default",vlx.volume)
        vlx.bar.add_mode("light",vlx.light)
        vlx.bar.init_window()

vol_lux_bar_demo = DemoVolLuxBar()



import threading
import colorama
from time import sleep
import tkinter as tk
from tkinter import ttk

def launch_gui():

    from volux import VoluxOperator
    from voluxcliprint import VoluxCliPrint
    from voluxlight import VoluxLight
    from voluxaudio import VoluxAudio
    from voluxgui import VoluxGui

    # inital values
    vis_added = False

    # create instance of VoluxOperator
    vlx = VoluxOperator()
    vlx.add_module(VoluxCliPrint())  # add VoluxCliPrint module
    vlx.add_module(VoluxAudio(sensitivity=8))  # add VoluxAudio module

    gui_shared_modules = [vlx.cli,vlx.audio]

    # try adding voluxlight module
    try:
        vlx.add_module(VoluxLight(instance_label="office",init_mode="group",init_mode_args={'group_label': 'Office'}))
        gui_shared_modules.append(vlx.light_office)
    except Exception as err:
        print("{}WARNING: couldn't add light module/s... reason: {}{}".format(colorama.Fore.YELLOW,err,colorama.Style.RESET_ALL))

    # try adding voluxlightvisualiser module
    try:
        from voluxlightvisualiser import VoluxLightVisualiser, INTENSE_MODE
        vlx.add_module(VoluxLightVisualiser(mode=INTENSE_MODE,packetHz=240,hueHz=240,hue_cycle_duration=5))
        gui_shared_modules.append(vlx.vis)
        vis_added = True
    except Exception as err:
        print("{}WARNING: couldn't add visualiser module... reason: {}{}".format(colorama.Fore.YELLOW,err,colorama.Style.RESET_ALL))

    vlx.add_module(VoluxGui(shared_modules=gui_shared_modules))

    if vis_added == True:
        visHz = 240
        def vis_func(loop_hz):
            while True:
                if vlx.gui.mainApp.vis_frame.vis_on.get() == True:
                    vlx.vis.start()
                    while vlx.gui.mainApp.vis_frame.vis_on.get() == True:
                        vlx.vis.set(vlx.audio.get())
                        vis_get = vlx.vis.get_from_gui()
                        vlx.light_office.set_color(
                            vis_get,
                            # duration=50
                        )
                        sleep(1/loop_hz)
                    vlx.vis.stop()

                        # print("START VISUALISER")
                        # vlx.vis.start()
                        # while vlx.vis._enabled == True:
                        #     vlx.vis.set(vlx.audio.get())
                        #     vlx.light.set_color(vlx.vis.get())
                        #     sleep(1/loop_hz)
                sleep(1)

        _vis_func_thread = threading.Thread(target=vis_func,args=(visHz,))
        _vis_func_thread.start()

    vlx.gui.init_window()
    vlx.audio.stop()

    if vis_added == True:
        vlx.vis.stop()



import threading
import numpy as np
import pyaudio
from time import sleep, process_time
from collections import deque
import lifxlan

from colorama import init

init()
from colorama import Fore, Back, Style

# ---- class-specific functions ----
# thanks to the commenter "Dima Tisnek" for this elegant solution to clamping:
# https://stackoverflow.com/q/4092528
clamp = lambda value, minv, maxv: max(min(value, maxv), minv)

# ---- functions ----


def print_bar(
    LightShow_Object, _name, _val, _seg_active, _seg_inactive, _total_segments=100
):

    # check inputed data
    if not (type(_seg_active) == BarSegment or type(_seg_inactive) == BarSegment):
        raise TypeError(
            "active_segment and inactive_segment must be BarSegment objects"
        )
    elif not (0 <= _val <= 1):  # ensure _val is between 0 and 1
        raise ValueError("val must be between 0 and 1, value was", _val)
    else:
        segments_activated = int(_val * _total_segments)

        active_string = _seg_active.gen_segments(segments_activated)
        inactive_string = _seg_inactive.gen_segments(
            _total_segments - segments_activated
        )

        if LightShow_Object.loop_time != None:
            target_Hz = LightShow_Object.packetHz
            actual_Hz = 1 / LightShow_Object.loop_time
            delta_Hz = int(actual_Hz - target_Hz)

            lrString = "({}Hz) ({:>3}Hz Δ) {}=[{}{}{}] ({:>3}) ({} devices)".format(
                target_Hz,
                delta_Hz,
                _name,
                active_string,
                inactive_string,
                Style.RESET_ALL,
                int(_val * 100),
                None
            )
            print(lrString)

        # Frequency domain representation


# adapted code from https://pythontic.com/visualization/signals/fouriertransform_fft
def return_FFT(amplitude, samplingFrequency):
    fourierTransform = np.fft.fft(amplitude) / len(amplitude)  # Normalize amplitude
    fourierTransform = fourierTransform[
        range(int(len(amplitude) / 2))
    ]  # Exclude sampling frequency

    tpCount = len(amplitude)
    values = np.arange(int(tpCount / 2))
    timePeriod = tpCount / samplingFrequency
    frequencies = values / timePeriod

    return (frequencies, abs(fourierTransform))


# ---- classes ----


class BarSegment:
    def __init__(self, symbol, fore="", back=""):
        self.symbol = symbol
        self.fore = fore
        self.back = back

    def gen_segments(self, amount):
        return self.fore + self.back + self.symbol * amount + Style.RESET_ALL


class LightShow:
    def __init__(
        self,
        mlifx,
        pyaudio,
        packetHz,
        fade,
        colorCycleHz=0.5,
        color_temp=6500,
        min_brightness=1,
        max_brightness=65535,
        min_saturation=0,
        max_saturation=65535,
        amp_sensitivity=9,
        initial_hue=0,
        hue_cycle_duration=5,
    ):

        self.mlifx = mlifx
        self.pyaudio = pyaudio
        self.packetHz = packetHz
        self.fade = fade
        self.color_temp = color_temp
        self.min_brightness = min_brightness
        self.max_brightness = max_brightness
        self.min_saturation = min_saturation
        self.max_saturation = max_saturation

        self.amp_sensitivity = amp_sensitivity

        self.pa = self.pyaudio.PyAudio()

        self.running = False
        self.threads = self._create_threads()
        self.stream = self._create_stream()
        self.audio_data = np.array([])
        self.audio_norm = 0
        self.audio_proportion = 0
        self.input_value = 0
        self.audio_hue_impact = 0.5  # float from 0 .. 1

        self.hue = initial_hue

        self.hue_rel = (
            0
        )  # relative hue (modified by external factors, isolated from normal hue)

        self.hue_min = 0
        self.hue_max = 65535
        self.hue_cycle_duration = hue_cycle_duration  # in seconds

        self.hue_increment = (self.hue_max / self.packetHz) / self.hue_cycle_duration

        self.needs_update = (
            False
        )  # make sure we don't run loops until data needed for those loops is ready

        self.loop_time = None

        self.mzl_colors = deque(self.mlifx.multizonelights[0].get_color_zones())

        self.set_colors = True

        self.pixel_color = (0,0,0,6500)

    def start(self):

        self.running = True  # enable threads
        self.mlifx.prepare()  # save state of all managed lights and power them on
        self._start_threads()
        self.stream.start_stream()

    def stop(self):

        self.running = False  # stop all threads
        self.mlifx.restore()  # restore original state of all managed lights
        self.stream.stop_stream()
        self.stream.close()
        self.pa.terminate()
        print(Style.RESET_ALL)  # reset ansi escapes

    def _create_threads(self):

        threads = {
            "set_loop": threading.Thread(target=self._set_loop),
            "hue_loop": threading.Thread(target=self._hue_loop),
        }

        return threads

    def _start_threads(self):

        for t in self.threads:
            self.threads[t].start()

    def _set_loop(self):

        while self.running == True:

            if self.needs_update == True:

                t1 = process_time()
                self.audio_norm = np.linalg.norm(self.audio_data) / len(self.audio_data)
                self.audio_proportion = clamp(self.audio_norm / 2 ** self.amp_sensitivity, 0, 1)  # normalized audio returning a value between 0 to 1

                # ---- SCRIPT PUT IN HACK-Y ----
                # data_frames = len(self.audio_data)
                # fft_frequencies, fourierTransform = return_FFT(self.audio_data, data_frames)
                # freq_count = 0
                # total_amp = 0
                # for i in range(len(fft_frequencies)):
                #     if (fft_frequencies[i] >= 0 and fft_frequencies[i] <= 60):
                #         freq_count += 1
                #         total_amp += fourierTransform[i]
                #     if (freq_count > 0):
                #         average_amp = total_amp / freq_count
                #     else:
                #         average_amp = 0
                # self.audio_proportion = average_amp/data_frames # todo: look into RMS (Root Mean
                #                                                 # Squared) to potentially fix spikes
                #                                                 # in bass while loud treble
                # self.audio_proportion = clamp(self.audio_proportion*self.amp_sensitivity,0,1)
                # ---- SCRIPT PUT IN HACK-Y ----

                self.hue_rel = self.hue + self.hue * np.sin(
                    self.input_value * self.audio_hue_impact
                )
                # self.hue_rel = self.hue

                self.hue_rel = (
                    self.hue_rel % self.hue_max
                )  # MAGIC PART THAT MAKES THE VALUE WRAP AROUND PERFECTLY

                h, s, v, k = self._intense_HSVK()

                self.pixel_color = (h, s, v/2, k)
                self.pixel_color2 = self._intense_HSVK()
                # self.pixel_color = (self.hue, 65535*1, max_brightness*self.audio_proportion, color_temp)
                pixel_empty = (0, 0, 0, k)

                # self.set_device_colors()

                # ==== ACTUALLY OKAY ====

                # i = 0 # index to start painting at
                # step = 0 # how many zones to paint with each color
                # iters = 1 # how many different zones to paint
                # self.mzl_colors.append(self.pixel_color)
                # self.mzl_colors.rotate(1)
                # for iter in range(iters):
                #     # print(iter)
                #     apply = 0
                #     if iter == iters - 1:
                #         apply = 1
                #     self.mlifx.multizonelights[0].set_zone_color(i+(step*iter), i+(step*iter)+step, self.mzl_colors[iter], duration=self.fade, rapid=True, apply=apply)

                # self.mzl_colors[i] = self.pixel_color
                # todo: experiment with even indexs as bass, odd indexes as treble


                    # sleep(0.020)

                # ==== ACTUALLY OKAY ====

                # self.mlifx.multizonelights[0].set_zone_color(i, i+step*1, self.mzl_colors[i], duration=self.fade, rapid=True, apply=0)
                # self.mlifx.multizonelights[0].set_zone_color(i+step*1, i+step*2, self.mzl_colors[i+2], duration=self.fade, rapid=True, apply=0)
                # self.mlifx.multizonelights[0].set_zone_color(i+step*2, i+step*3, self.mzl_colors[i+4], duration=self.fade, rapid=True, apply=0)
                # self.mlifx.multizonelights[0].set_zone_color(i+step*3, i+step*4, self.mzl_colors[i+6], duration=self.fade, rapid=True, apply=1)

                # self.mlifx.multizonelights[0].set_zone_color(i+12, i+15, self.mzl_colors[i+8], duration=self.fade, rapid=True, apply=1)

                # self.mlifx.multizonelights[0].set_zone_color(1, 1, self.mzl_colors[1], duration=self.fade, rapid=True, apply=0)
                # self.mlifx.multizonelights[0].set_zone_color(2, 2, self.mzl_colors[2], duration=self.fade, rapid=True, apply=0)
                # self.mlifx.multizonelights[0].set_zone_color(3, 3, self.mzl_colors[3], duration=self.fade, rapid=True, apply=1)
                # self.mlifx.multizonelights[0].set_zone_color(4, 4, self.mzl_colors[4], duration=self.fade, rapid=True, apply=0)
                # for i in range(len(self.mzl_colors)):
                #     if i == len(self.mzl_colors):
                #         apply = 1
                #     else:
                #         apply = 0
                #     self.mlifx.multizonelights[0].set_zone_color(0, 0, self.mzl_colors[0], duration=self.fade, rapid=True)
                # self.mlifx.mutlizonelights[1].set

                # for zone in range(len(self.mzl_colors)):
                # self.mlifx.multizonelights[0].set_zone_color(self.mzl_colors, duration=self.fade, rapid=True)
                # sleep(0.1)
                # for i in range(len(mzl_colors)):
                #     mzl.set_zone_color(i, i+1, self.pixel_color, duration=self.fade, rapid=True)
                # mzl_colors = np.roll(mzl_colors, 1)
                # mzl.set_zone_colors(mzl_colors, duration=self.fade, rapid=True)
                # mzl.set_zone_color(0, 7, self.pixel_color, duration=self.fade, rapid=True)

                 # mzl.set_zone_color(7, 7, self.pixel_color, duration=self.fade, rapid=True)
                # mzl.set_zone_color(1, 1, [c*0.75 for c in self.pixel_color], duration=self.fade, rapid=True)
                # mzl.set_zone_color(6, 6, [c*0.75 for c in self.pixel_color], duration=self.fade, rapid=True)

                self.needs_update = False

                t2 = process_time()

                while (
                    t2 - t1 < 1 / self.packetHz
                ):  # loop until we've waited as long as the packetHz specifies
                    t2 = process_time()

                self.loop_time = t2 - t1

    def _hue_loop(self):

        while self.running == True:

            self.hue = self.hue + self.hue_increment  # add increment

            if self.hue > self.hue_max:
                self.hue = self.hue - self.hue_max
            elif self.hue < self.hue_min:
                self.hue = self.hue + self.hue_max

            sleep(1 / self.packetHz)

    def set_device_colors(self):
        if self.set_colors == True:

            # for mlight in self.mlifx.managed_lights:
            #     mlight.light.set_color(self.pixel_color, self.fade, rapid=True)
            #
            # for m_tc in self.mlifx.managed_tilechains:
            #     m_tc.canvas = np.roll(
            #         m_tc.canvas, 1, axis=1
            #     )  # shift canvas on axis
            #     m_tc.canvas = np.roll(
            #         m_tc.canvas, 1, axis=0
            #     )  # shift canvas on axis
            #     m_tc.paint_line(self.pixel_color, "y", 0)  # paint line on side of canvas
            #     m_tc.paint_line(self.pixel_color, "x", 0)  # paint line on side of canvas
            #     m_tc.paint_line(self.pixel_color, "y", 7)  # paint line on side of canvas
            #     m_tc.paint_line(self.pixel_color, "x", 7)  # paint line on side of canvas
            #
            #     m_tc.update_tilechain()

            # self.mlifx.multizonelights[0].set_color(self.pixel_color, self.fade, rapid=True)

            # self.mzl_colors = np.roll(self.mzl_colors, 1, axis=0)
            # self.mzl_colors[0] = self.pixel_color

            for mzl in self.mlifx.multizonelights:
                mzl.set_color(self.pixel_color2, self.fade, rapid=True)

    def _intense_HSVK(self):

        h = self.hue_rel
        s = (
            self.min_saturation
            + (self.max_saturation - self.min_saturation) * self.input_value
        )  # regular saturation - higher levels = higher saturation
        v = (
            self.min_brightness
            + (self.max_brightness - self.min_brightness) * self.input_value
        )
        k = self.color_temp

        return (h, s, v, k)

    def _whiteout_HSVK(self):

        h = self.hue_rel
        s = self.max_saturation * np.cos(
            self.min_saturation - self.input_value
        )  # inverse saturation - higher levels = lower saturation
        v = (
            self.min_brightness
            + (self.max_brightness - self.min_brightness) * self.input_value
        )
        k = self.color_temp

        return (h, s, v, k)

    def _blackout_HSVK(self):

        h = self.hue_rel
        s = (
            self.min_saturation
            + (self.max_saturation - self.min_saturation) * self.input_value
        )  # regular saturation - higher levels = higher saturation
        v = (
            self.max_brightness
            - (self.max_brightness - self.min_brightness) * self.input_value
        )
        k = self.color_temp

        return (h, s, v, k)

    def _just_red_HSVK(self):

        h = 0
        s = (
            self.min_saturation
            + (self.max_saturation - self.min_saturation) * self.input_value
        )  # regular saturation - higher levels = higher saturation
        v = (
            self.min_brightness
            + (self.max_brightness - self.min_brightness) * self.input_value
        )
        k = self.color_temp

        return (h, s, v, k)



class VoluxDemo:
    def __init__(self, demo_name, demo_method, alias, requirements=None, *args, **kwargs):
        self._name = demo_name
        self._method = demo_method
        self._alias = alias
        self._requirements = requirements # optional

    def run(self):
        self._method()

    # def __str__(self):
    #     header = "[DEMO:{name}]----".format(name=self._alias)
    #     divider = "-"*len(header)
    #     body = "Name: {name}\nAlias: {alias}\nRequirements: {requirements}".format(name=self._name,requirements=self._requirements, alias=self._alias)
    #
    #     return("{header}\n{body}\n{divider}".format(header=header,divider=divider,body=body))

    def __repr__(self):
        return "<VoluxDemo '{}'>".format(self._alias)



class MixerController:

    def __init__(self):

        import alsaaudio
        try:
            # print("mixers:",alsaaudio.mixers())
            # print("PCM:",alsaaudio.PCM().cardname())
            self.mixer = alsaaudio.Mixer("Master")
        except FileNotFoundError:
            print("I see you're running on a CI server -.-")

    def svol(self,newvol):
        if type(newvol) == int:

            if (newvol >= 0) and (newvol <= 100):
                try:
                    self.mixer.setvolume(newvol)
                except:
                    raise Exception
            else:
                raise ValueError("volume must be between 0 and 100")
        else:
            raise TypeError("volume must be an integer")

    def gvol(self):

        _gvol = self.mixer.getvolume()
        return(int(sum(_gvol) / len(_gvol))) # return average across channels

    def smute(self,newstate):

        if type(newstate) == bool:

            if newstate == False:
                self.mixer.setmute(0)
            elif newstate == True:
                self.mixer.setmute(1)
            else:
                raise Exception("WHAT?")

        else:
            raise TypeError("Input must be boolean")

    def gmute(self):

        _gmute = self.mixer.getmute()
        mute_average = int(sum(_gmute) / len(_gmute))

        if mute_average == 0:
            return(False)
        elif mute_average == 1:
            return(True)

        else:
            raise Exception("mute_average didn't return 0 or 1, please create an issue on Github!")



# determine the platform

import platform
platform = platform.system()

# load the necessary classes / packages / whatever

if platform == "Windows":
    from cpaudio.Platforms.Windows import MixerController

elif platform == "Darwin":
    from cpaudio.Platforms.Darwin import MixerController

elif platform == "Linux":
    from cpaudio.Platforms.Linux import MixerController

else:
    raise Exception("Platform not recognised!")

# create common aliases for them all regardless of platform
    # tool 1
    # tool 2
    # tool 3
    # etc.


from volux import VoluxModule


class VoluxCliPrint(VoluxModule):
    def __init__(self, *args, **kwargs):
        super().__init__(
            module_name="Volux CLI Print",
            module_attr="cli",
            module_get=self.get,
            get_type=int,
            get_min=0,
            get_max=100,
            module_set=self.set,
            set_type=int,
            set_min=0,
            set_max=100,
            shared_modules=[],
            pollrate=None
        )
        self.cli_val = 0

    def get(self):

        return self.cli_val

    def set(self, new_val):

        self.cli_val = new_val
        print("VoluxCliPrint new value: {}".format(self.cli_val))



import numpy as np
import lifxlan

# ---- classes ----
class HSVKPixel:
    def __init__(self, hsvk_tuple):

        self.h, self.s, self.v, self.k = hsvk_tuple

    def write(self, hsvk_tuple):

        self.h, self.s, self.v, self.k = hsvk_tuple

    def read(self):

        return (self.h, self.s, self.v, self.k)


class ManagedTilechain:
    def __init__(self, TileChain_Object):
        self.TileChain = TileChain_Object
        self.size = TileChain_Object.get_canvas_dimensions()
        self.size_all_values = (self.size[0], self.size[1], 4)
        self.num_tiles = TileChain_Object.get_tile_count()
        self.num_pixels = ((self.size[0] + 1) * (self.size[1] + 1) * 4) - 1
        self.off_pixel = (0, 0, 0, 6500)
        self.canvas = self._gen_empty_frame()

        self.power = None
        self.color = None

    def _gen_empty_frame(self):

        empty_frame = np.full(self.size_all_values, self.off_pixel, dtype=object)

        return empty_frame

    def read_HSVK_2D(self):

        HSVK_list = []

        # for each cell of self.canvas
        for y in self.canvas:
            for x in y:
                # store the output of cell.read() in 1D array
                HSVK_list.append(x)

        # convert the 1D array to 2D (the third dimension is HSVK data)
        HSVK_2D = np.reshape(HSVK_list, (self.size[0], self.size[1], 4))

        return HSVK_2D

    def read_HSVK_tiles(self):

        # convert the 1D array to a 2D array of 64 pixels per tile
        HSVK_tiles = np.reshape(self.canvas, (self.num_tiles, 64, 4))

        return HSVK_tiles

    def print_HSVK_2D(self):

        # print data relating to each pixel in the HSVK_2D array
        HSVK_2D = self.read_HSVK_2D()
        n = 0
        for y in range(len(HSVK_2D)):
            for x in range(len(HSVK_2D[y])):
                print(
                    "pixel:{} x-index:{} y-index:{} color:{}".format(
                        n, x, y, HSVK_2D[y][x]
                    )
                )
                n += 1

    def print_HSVK_tiles(self):

        # print 2D tile pixels array
        HSVK_tiles = self.read_HSVK_tiles()
        for tile in range(len(HSVK_tiles)):
            for pixel in range(len(HSVK_tiles[tile])):
                print(
                    "tile:{} pixel:{} HSVK:{}".format(
                        tile, pixel, HSVK_tiles[tile][pixel]
                    )
                )

    def update_tilechain(self, fade=0, rapid=True):

        HSVK_tiles = self.read_HSVK_tiles()
        self.TileChain.set_tilechain_colors(HSVK_tiles, fade, rapid=rapid)

    def paint_pixel(self, color, x, y):

        self.canvas[y][x] = color

    def paint_line(self, color, dimension, index):

        if dimension == "x":
            for x in range(self.size[0]):
                self.canvas[index][x] = color

        elif dimension == "y":
            for y in range(len(self.canvas)):
                self.canvas[y][index] = color

        else:
            raise Exception(
                "dimension must be either the string 'x' or 'y', not {}".format(
                    str(dimension)
                )
            )

    def ssave(self):

        self.power = self.TileChain.get_power()
        self.tilechain_colors = self.TileChain.get_tilechain_colors()

    def sload(self):

        self.TileChain.set_power(self.power)
        self.TileChain.set_tilechain_colors(self.tilechain_colors)


# ---- script ----
if __name__ == "__main__":
    lifx = lifxlan.LifxLAN()
    tilechains = lifx.get_tilechain_lights()
    tilechain = tilechains[0]

    my_mtc = ManagedTilechain(tilechain)
    HSVK_list = my_mtc.read_HSVK_2D()
    print(HSVK_list)

    tilechain.set_tilechain_colors(HSVK_list, 0, rapid=True)



class UnsupportedFeature:
    def __init__(self):
        pass

class DeviceState:
    def __init__(self,power=None,color=None,infrared=None):
        self.power = power
        self.color = color

class ManagedDevice:
    def __init__(self,device):
        self.device = device
        self.label = device.get_label()
        self.power = device.get_power()
        self.is_light = device.is_light()
        self.supports_color = device.supports_color()

        self.color = None
        # self.temperature = UnsupportedFeature()
        # self.multizone = UnsupportedFeature()
        # self.infrared = UnsupportedFeature()

    def ssave(self):
        """save device state inside of class"""
        self.power = self.device.get_power()

        if (self.supports_color == True): self.color = self.device.get_color()
        elif (self.supports_color == False): self.color = UnsupportedFeature()
        else: raise TypeError("self.supports_color should be a boolean value")

    def sload(self):
        """load device state last saved inside of class"""
        self.device.set_power(self.power)

        if (self.supports_color == True):
            self.device.set_color(self.color)


# class ManagedLight:
#     def __init__(self, _light, debug=False):
#         if debug == True:
#             print("creating a new ManagedLight...")
#         self.light = _light
#         self.power = None
#         self.color = None
#         self.infrared = None
#
#     def ssave(self):
#         """save light state inside of class"""
#         self.power = self.light.get_power()
#         self.color = self.light.get_color()
#         # self.infrared = self.light.get_infrared()
#
#     def sload(self):
#         """load light state last saved inside of class"""
#         self.light.set_color(self.color)
#         self.light.set_power(self.power)
#         # self.light.set_infrared(self.infrared)
#
#     def sexport(self):
#         """export light state to outside class"""
#         return DeviceState(power=self.power,color=self.color,infrared=self.infrared)
#
#     def simport(self,devicestate):
#         """import light state from outside class"""
#         self.power, self.color, self.infrared = devicestate.power, devicestate.color, devicestate.infrared
#
#     def print_saved_state(self):
#         """print the saved state of the light"""
#         print(
#             "[{} (saved)] power:{} color:{} infrared:{}".format(
#                 self.light.get_label(), self.power, self.color, self.infrared
#             )
#         )



# function by Denver P. 2019
from colorama import init
init()
from colorama import Fore, Back, Style

def hsv2ansi(h,s,v):
    ansi = None
    red_to_yellow = 45/360
    # yellow_to_green = 150/360 # more accurate to reality, doesn't quite match lights
    yellow_to_green = 100/360
    green_to_cyan = 195/360
    cyan_to_blue = 225/360
    blue_to_magenta = 285/360
    magenta_to_red = 345/360

    # get colour based on hue
    if (h>=magenta_to_red or h<red_to_yellow):
        ansi_fore = Fore.RED
        ansi_back = Back.RED
    elif (h>=red_to_yellow and h<yellow_to_green):
        ansi_fore = Fore.YELLOW
        ansi_back = Back.YELLOW
    elif (h>=yellow_to_green and h<green_to_cyan):
        ansi_fore = Fore.GREEN
        ansi_back = Back.GREEN
    elif (h>=green_to_cyan and h<cyan_to_blue):
        ansi_fore = Fore.CYAN
        ansi_back = Back.CYAN
    elif (h>=cyan_to_blue and h<blue_to_magenta):
        ansi_fore = Fore.BLUE
        ansi_back = Back.BLUE
    elif (h>=blue_to_magenta and h<magenta_to_red):
        ansi_fore = Fore.MAGENTA
        ansi_back = Back.MAGENTA
    else:
        print("I DUNNO! D:")
        raise Exception("color doesn't fit any conditions!")

    return(ansi_fore,ansi_back)

def demo():
    for i in range(360):
        ansi = hsv2ansi(i/360,0,0)
        print("{}° = {}demo{}".format(i,ansi[0],Style.RESET_ALL))

if __name__ == '__main__':
    demo()



from cpaudio import MixerController
from time import sleep

mixer = MixerController()

old_vol = mixer.gvol()
old_mute_state = mixer.gmute()

print("old volume: {}".format(old_vol))
print("old mute state: {}".format(old_mute_state))

print(); sleep(1)

print("setting volume to 80%")
mixer.svol(80)
print("volume reported: {}".format(mixer.gvol()))

print(); sleep(1)

print("setting mute...")
mixer.smute(True)
print("mute state reported: {}".format(mixer.gmute()))

print(); sleep(1)

print("reverting to original values...")

print(); sleep(1)

mixer.svol(old_vol)
mixer.smute(old_mute_state)
print("volume reported: {}".format(mixer.gvol()))
print("mute state reported: {}".format(mixer.gmute()))

print(); sleep(1)

print("finished!")



from voluxaudio.lifxtools import DeviceState

class ManagedLight:
    def __init__(self, _light, debug=False):
        if debug == True:
            print("creating a new ManagedLight...")
        self.light = _light
        self.power = None
        self.color = None
        self.infrared = None

    def ssave(self):
        """save light state inside of class"""
        self.power = self.light.get_power()
        self.color = self.light.get_color()
        # self.infrared = self.light.get_infrared()

    def sload(self):
        """load light state last saved inside of class"""
        self.light.set_color(self.color)
        self.light.set_power(self.power)
        # self.light.set_infrared(self.infrared)

    def sexport(self):
        """export light state to outside class"""
        return DeviceState(power=self.power,color=self.color,infrared=self.infrared)

    def simport(self,devicestate):
        """import light state from outside class"""
        self.power, self.color, self.infrared = devicestate.power, devicestate.color, devicestate.infrared

    def print_saved_state(self):
        """print the saved state of the light"""
        print(
            "[{} (saved)] power:{} color:{} infrared:{}".format(
                self.light.get_label(), self.power, self.color, self.infrared
            )
        )



"""
lifxtools package
"""

print("lifxtools/__init__.py")

from .manageddevice import *
from .virtualdevice import *
from .managedlifx import *
from .managedlight import *
from .managedtilechain import *
from .lightshow import *
from .color_presets import *

# from .message import *
# from .msgtypes import *
# from .unpack import unpack_lifx_message
# from .device import *
# from .light import *
# from .multizonelight import *
# from .group import Group
# from .tilechain import TileChain, Tile
# from .utils import *

# __version__     = '0.0.1'
# __description__ = 'A set of tools for controling LIFX devices on the local network over lan'
# __url__         = 'http://github.com/drtexx/lifxtools'
# __author__      = 'Denver Pallis'
# __license__     = 'GPLv3+'


# imports
from time import sleep, process_time  # used for delays and benchmarking
from lifxlan import LifxLAN, RED, WHITE  # used for controlling lights
from colorsys import rgb_to_hsv, hsv_to_rgb

# settings
num_lights = None  # makes discovery much faster when specified instead of none
live_data = True
debug = True

# decorators
def d_benchmark(func):
    def func_wrapper(*args, **kwargs):
        t1 = process_time()  # take first snapshot of processing time
        result = func(*args, **kwargs)
        t2 = process_time()  # take second snapshot of processing time
        loop_time = t2 - t1
        print(
            "[{}] ({:>3} Hz) took {} seconds to complete!".format(
                func.__name__, int(1 / (loop_time)), loop_time
            )
        )
        return result

    return func_wrapper


def d_debug_messages(func):
    def func_wrapper(*args, **kwargs):
        if debug == True:
            print("[{}] started...".format(func.__name__))
        try:
            return func(*args, **kwargs)
        except Exception as err:
            print("[{}] ERROR!: {}".format(func.__name__, err))
        finally:
            if debug == True:
                print("[{}] finished!".format(func.__name__))

    return func_wrapper


# functions
@d_debug_messages
def return_interface(num_lights):

    if num_lights == None:

        return LifxLAN()

    elif num_lights > 0:

        print(
            "WARNING: num_lights is not None. Make sure it is set to your actual number of devices or you will likely have issues!"
        )
        return LifxLAN(num_lights)

    else:

        raise TypeError("num_lights must be an interger greater than 0")


def return_num_lights(devices):
    if num_lights != None:
        return num_lights
    else:
        return len(devices)


@d_debug_messages
def list_devices(devices):
    i = 0
    for device in devices:
        if debug == True:
            print(
                "devices[{}] = [label='{}', power={}, color={}])".format(
                    i, device.get_label(), device.get_power(), device.get_color()
                )
            )
        i += 1


def blink_devices(devices):
    """ blink all devices found one-by-one """
    for device in devices:
        original_power = device.get_power()

        device.set_power(False, 0.1)
        sleep(0.5)
        device.set_power(True, 0.1)
        sleep(0.5)

        device.set_power(original_power)


def toggle_light(_light):
    light_power = _light.get_power()
    if light_power == 0:
        _light.set_power(True)
        if debug == True:
            print("{} turned on".format(_light.get_label()))
    elif light_power > 0:
        _light.set_power(False)
        if debug == True:
            print("{} turned off".format(_light.get_label()))
    else:
        print(
            "WIP: power other than True or False not currently supported, using 65535 range is not yet implemented"
        )


def set_light_color(_light, color):
    _light.set_color(color)
    if debug == True:
        print("{} color set to {}".format(_light.get_label(), color))


@d_debug_messages
def list_lights(_lights):
    for light in _lights:
        print(
            "[{}] power:{} color:{} infrared:{}".format(
                light.get_label(),
                light.get_power(),
                light.get_color(),
                light.get_infrared(),
            )
        )

def blink_light(_light, delay=1):
    _light.set_power(False, rapid=True)
    sleep(delay)
    _light.set_power(True, rapid=True)


def get_lights(_interface, debug=False):

    try:
        if debug == True:
            print("[ get lights ] started...")
        return _interface.get_lights()

    except:
        if debug == True:
            print("[ get lights ] ERROR!")

    finally:
        if debug == True:
            print("[ get lights ] finished!")


def create_managed_lights(_lights):
    _ManagedLights = []
    for light in _lights:
        _ManagedLights.append(ManagedLight(light))
    return _ManagedLights


def rgbk2hsvk(r, g, b, k):
    """ convert rgb + kelvin to hsvk for bulbs (colors conversion/scaling) """
    h, s, v = rgb_to_hsv(r / 255.0, g / 255.0, b / 255.0)
    h = h * 0xFFFF
    s = s * 0xFFFF
    v = v * 0xFFFF
    k = k
    return (h, s, v, k)

def hsv2rgb(h, s, v):
    r, g, b = hsv_to_rgb(h/65535, s/65535, v/65535)
    return (r, g, b)

def prepare_ManagedLights(_ManagedLights):
    for ml in _ManagedLights:
        ml.ssave()
        ml.light.set_power(True)

def restore_ManagedLights(_ManagedLights):
    for ml in _ManagedLights:
        ml.sload()



#!/usr/bin/env python -e

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# personal note: to refresh auto-documentation
# (navigate to repo root first)
# sphinx-apidoc -o docs/modules . test_* cpaudio setup.py voluxdisplay/cpdisplay

# how-to metadata
# https://packaging.python.org/specifications/core-metadata/

import volux  # CHANGE THIS TO MATCH YOUR PACKAGE'S NAME!

mypackage = volux

readme_filename = 'README.md'
readme_encoding = 'utf-8'
long_description_content_type = 'text/markdown'

import setuptools # Always prefer setuptools over distutils
from codecs import open # To use a consistent encoding
from os import path
here = path.abspath(path.dirname(__file__))
with open(path.join(here, readme_filename), encoding=readme_encoding) as f:
    long_description = f.read() # Get the long description from the README file

setuptools.setup(
    metadata_version=mypackage.__metadata_version__, # Optional
    name=mypackage.__name__, # Required
    version=mypackage.__version__, # Required
    summary=mypackage.__summary__, # Optional
    description=mypackage.__description__, # Optional
    description_content_type=mypackage.__description_content_type__,  # Optional
    keywords=mypackage.__keywords__,
    url=mypackage.__homepage__,
    author=mypackage.__author__,
    author_email=mypackage.__author_email__,
    license=mypackage.__license__,
    classifiers=mypackage.__classifiers__,
    install_requires=mypackage.__requires_dist__,
    requires_python=mypackage.__requires_python__,
    requires_external=mypackage.__requires_external__,
    project_urls=mypackage.__project_urls__,

    long_description=long_description, # Optional
    long_description_content_type=long_description_content_type, # Optional

    package_data={ # Optional
        'sample': ['package_data.dat'],
    },
    entry_points={ # Optional
        'console_scripts': [
            '{}={}:__main__.main'.format(mypackage.__name__,mypackage.__name__),
        ],
    },
    packages=setuptools.find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
)



default_color = (58275, 0, 33968, 3500)  # roughly 50% brightness at 3500k
bedtime_color = (58275, 0, 18010, 2000)  # roughly 25% brightness at 2000k
theatre_color = (58275, 0, 1, 2000)  # minimum brightness at 2000k



import lifxlan
from voluxaudio import lifxtools
from .managedtilechain import ManagedTilechain


class ManagedLifx:
    def __init__(self, lifx, verbose=False, create_managed=True):

        self.lifx = lifx
        self.verbose = verbose

        self.devices, self.lights, self.tilechains, self.multizonelights = [], [], [], []
        self.managed_devices, self.managed_lights, self.managed_tilechains, self.managed_multizonelights = [], [], [], []
        self.device_count, self.light_count, self.tilechain_count = None, None, None

        if self.verbose == True:
            print("discovering LIFX devices...")
        self.refresh_devices(refresh_managed=create_managed)

    # refresh list of devices and (unless specified not to) sort the list accordingly
    def refresh_devices(self, sort=True, refresh_managed=True):

        self.devices = self.lifx.get_devices()

        if sort == True:

            self._sort_devices()

        if refresh_managed == True:

            self.remanage_devices()

    def remanage_devices(self):

        self._refresh_managed_devices()
        self._refresh_managed_lights()
        self._refresh_managed_tilechains()

    def _refresh_managed_devices(self):

        self.managed_devices = []

        for device in self.devices:

            self.managed_devices.append(lifxtools.ManagedDevice(device))

    def _refresh_managed_lights(self):

        self.managed_lights = []

        for light in self.lights:

            self.managed_lights.append(lifxtools.ManagedLight(light))

    def _refresh_managed_tilechains(self):

        self.managed_tilechains = []

        for tilechain in self.tilechains:

            self.managed_tilechains.append(ManagedTilechain(tilechain))

    def _refresh_managed_multizonelights(self):

        self.managed_multizonelights = []

        for device in self.managed_devices:
            if type(device) == lifxlan.MultiZoneLight:
                self.managed_multizonelights.append(device)

    def print_device_labels(self):

        for device in self.devices:

            print(device.get_label())

    # put devices in respective lists, e.g. TileChain types in .tilechains, etc.
    def _sort_devices(self, get_counts=True):

        for device in self.devices:

            device_type = type(device)

            if device_type == lifxlan.light.Light:

                self.lights.append(device)

            elif device_type == lifxlan.tilechain.TileChain:

                self.tilechains.append(device)

            elif device_type == lifxlan.multizonelight.MultiZoneLight:

                self.multizonelights.append(device)

            else:

                raise TypeError(
                    "type of",
                    device.get_label(),
                    "is",
                    device_type,
                    "and we don't recognize it!",
                )

        if self.verbose == True:
            print("finished sorting devices!")
        if self.verbose == True:
            self.print_sorted_lists()

        if get_counts == True:
            self._get_device_counts()

    def print_sorted_lists(self):

        print("devices", self.devices)
        print("lights", self.lights)
        print("tilechains", self.tilechains)

    def _get_device_counts(self):

        self.device_count = len(self.devices)
        self.light_count = len(self.lights)
        self.tilechain_count = len(self.tilechains)

    def add_device(self,new_device):

        self.devices.append(new_device)
        self.managed_devices.append(lifxtools.ManagedDevice(new_device))

    def prepare(self):

        for mlight in self.managed_lights:

            mlight.ssave()
            mlight.light.set_power(True)

        for mtilechain in self.managed_tilechains:

            mtilechain.ssave()
            mtilechain.TileChain.set_power(True)

    def restore(self):

        for mlight in self.managed_lights:

            mlight.sload()

        for mtilechain in self.managed_tilechains:

            mtilechain.sload()



from volux import VoluxModule
import lifxlan


class VoluxLight(VoluxModule):
    def __init__(self, instance_label, init_mode, init_mode_args=[], group=None, shared_modules=[], pollrate=None, *args, **kwargs):
        super().__init__(
            module_name="Volux Light ({})".format(instance_label),
            module_attr="light_{}".format(instance_label),
            module_get=self.get,
            get_type=float,
            get_min=0,
            get_max=65535,
            module_set=self.set,
            set_type=float,
            set_min=0,
            set_max=65535,
            shared_modules=shared_modules,
            pollrate=pollrate
        )
        init_mode_options = ["all_devices","device","group"]  # note: all types of labels are caps-sensitive
        self.instance_label = instance_label
        self.init_mode = init_mode
        self.init_mode_args = init_mode_args
        # self.group = group  # note: group labels are caps sensitive

        self.lifx = lifxlan.LifxLAN(None)

        self.devices = []

        if not self.init_mode in init_mode_options:
            raise ValueError("invalid init_mode. options include: {}".format(init_mode_options))
        else:
            if self.init_mode == "all_devices":
                self.devices = self.lifx.get_devices()

            elif self.init_mode == "device":
                if type(self.init_mode_args) == dict:
                    if 'ip' in self.init_mode_args and 'mac' in self.init_mode_args:
                        raise NotImplementedError()  # note: implement ip + mac option
                    elif 'label' in self.init_mode_args:
                        self.devices.append(self.lifx.get_device_by_name(self.init_mode_args['label']))
                    else:
                        raise KeyError("'label' not specified in mode_args")
                else:
                    raise TypeError("mode_args must be of type 'dict'")

            elif self.init_mode == "group":
                if 'group_label' in self.init_mode_args:
                    devices = self.lifx.get_devices_by_group(self.init_mode_args['group_label']).devices
                    if len(devices) > 0:
                        self.devices = devices
                    else:
                        raise Exception("no devices found for group_label {}! Please note group_label is caps-sensitive!".format(self.init_mode_args['group_label']))
                else:
                    raise KeyError("'group_label' not specified in mode_args")

        # if not self.group == None:
        #     for device in self.devices:
        #         if not device.get_group_label() == self.group:
        #             device.set_power(False)
        #             self.devices.remove(device)
        #             print("removed '{}' from selected devices because it's not in group '{}'".format(device.get_label(),self.group))

        for device in self.devices:
            device.set_power(True)

        print("devices: {}".format(self.devices))

    def get(self):

        for device in self.devices:

            color = device.get_color()
            power = color[2] / 65535
            return power * 100

    def set(self, new_val):

        if new_val < 0:
            new_val = 0

        elif new_val > 100:
            new_val = 100

        new_val = new_val / 100

        for device in self.devices:

            color = device.get_color()
            new_color = (color[0], color[1], new_val * 65535, color[3])
            device.set_color(new_color)

    def set_color(self, new_color, duration=20, rapid=True):

        for device in self.devices:

            device.set_color(new_color, duration=duration, rapid=rapid)

    def toggle(self):

        for device in self.devices:

            power = device.get_power()
            device.set_power(not power)  # not ideal behaviour if all lights in different states



from volux import VoluxModule
import threading
import numpy as np
import pyaudio
from time import process_time, sleep
import tkinter as tk
from tkinter import ttk


class VoluxLightVisualiser(VoluxModule):
    def __init__(self, mode, packetHz, hueHz,  hue_cycle_duration,
    shared_modules=[], pollrate=100, initial_input_value=0, hue_min=0,
    hue_max=65535, colortemp=6500, min_saturation=0, max_saturation=65535,
    *args, **kwargs):
        super().__init__(
            module_name="Volux Light Visualiser",
            module_attr="vis",
            module_get=self.get,
            get_type=tuple,
            get_min=(0,0,0,0),
            get_max=(65535,65535,65535,9000),
            module_set=self.set,
            set_type=float,
            set_min=0,
            set_max=100,
            shared_modules=shared_modules,
            pollrate=pollrate
        )
        self.tk_frame = VoluxLightVisualiser_TkFrame
        self.frame_attr = 'vis_frame'
        # self.gui = None  # must be set via _set_gui_instace once self.tk_frame has been initialised elsewhere
        self.input_value = initial_input_value
        self.colortemp = colortemp

        self.minsat = min_saturation
        self.maxsat = max_saturation

        self.mode = mode
        self.packetHz = packetHz  # how often to send network packets to devices
        self.hueHz = hueHz  # how often to increment the hue
        self.hue_min = hue_min
        self.hue_max = hue_max
        self.hue_cycle_duration = hue_cycle_duration

        self._enabled = False  # initial value for _enabled
        self._hue = 0  # initial value for _hue
        self._rel_hue = self._hue  # initial value for _rel_hue

        self.threads = [
            threading.Thread(
                target=self._hue_cycle_func,
                args=(
                    self._get_hue_increment(),
                    1 / self.hueHz,
                    65535/2
                )
            )
        ]

    def get(self):

        return self.mode.get_vis_color(
            input_hue=self._rel_hue,
            input_colortemp=self.colortemp,
            input_value=self.input_value,
            minsat=self.minsat,
            maxsat=self.maxsat,
            minbri=0,
            maxbri=65535
        )

    def get_from_gui(self):

        return self.mode.get_vis_color(
            input_hue=self._rel_hue,
            input_colortemp=self.colortemp,
            input_value=self.input_value,
            minsat=self.gui.mainApp.vis_frame.minsat.get(),
            maxsat=self.gui.mainApp.vis_frame.maxsat.get(),
            minbri=0,
            maxbri=65535
        )

    def set(self, new_val):

        self.input_value = new_val  # 0 .. 100

    def start(self):

        if self._enabled == True:

            raise Exception("already enabled!")

        elif self._enabled == False:

            self._enabled = True
            self._start_threads()

        else:

            raise Exception("'enabled' should be bool")

    def stop(self):

        self._enabled = False

    def _start_threads(self):

        for thread in self.threads:
            thread.start()

    def _get_hue_increment(self):

        return (self.hue_max / self.hueHz) / self.hue_cycle_duration

    def _hue_cycle_func(self,hue_increment,time_per_hue_step,input_value_impact):

        while self._enabled == True:

            new_val = self._hue + hue_increment

            rel_input_multiplier = (self.input_value/100)*input_value_impact
            rel_hue = self._hue + rel_input_multiplier

            self._hue = new_val % self.hue_max  # limit the value to hue_max
            self._rel_hue = rel_hue % self.hue_max

            # print("_hue:",self._hue)
            cli_bar_raw_perc = int((self._hue / self.hue_max) * 100)
            cli_bar_rel_perc = int((self._rel_hue / self.hue_max) * 100)
            cli_bar_diff = cli_bar_rel_perc - cli_bar_raw_perc
            full_bar = "#"*cli_bar_raw_perc + "@"*cli_bar_diff
            print("[{:<100}]".format(full_bar))
            # print("[{:<100}]".format("#"*cli_bar_perc))
            # print("_rel_hue:",self._rel_hue)


            sleep(time_per_hue_step)

    def _get_rel_hue(self):

        pass

    def _set_gui_instance(self, gui):

        setattr(self,'gui',gui)

class VisMode:
    def __init__(self,vis_color_method,*args,**kwargs):
        self.get_vis_color = vis_color_method

class IntenseMode(VisMode):
    def __init__(self,*args,**kwargs):
        super().__init__(
            vis_color_method=self.get_color_method,
        )

    def get_color_method(self,input_hue,input_colortemp,input_value,minsat,maxsat,minbri,maxbri):

        # print("---get vis color wip---")
        # print("input_hue:",input_hue)
        # print("input_colortemp:",input_colortemp)
        # print("input_value:",input_value)
        # print("input_value/100:",input_value/100)
        # print("minsat:",minsat)
        # print("maxsat:",maxsat)
        # print("minbri:",minbri)
        # print("maxbri:",maxbri)

        satrange = (maxsat - minsat)
        brirange = (maxbri - minbri)

        # print("satrange:",satrange)
        # print("brirange:",brirange)

        h = input_hue
        s = minsat + (satrange * (input_value/100))  # regular saturation - higher levels = higher saturation
        b = minbri + (brirange * (input_value/100))
        k = input_colortemp

        # print((h,s,b,k))

        return (h, s, b, k)

INTENSE_MODE = IntenseMode()

class VoluxLightVisualiser_TkFrame(ttk.Frame):
    def __init__(self,parent,*args,**kwargs):
        super().__init__(parent,*args,**kwargs)
        self.parent = parent
        self.vis_on = tk.BooleanVar()
        self.minsat = tk.DoubleVar()
        self.maxsat = tk.DoubleVar()

        self.visualiser_title = ttk.Label(self, text="VISUALISER")
        self.visualiser_title.pack()

        self.visualiser_checkbox = ttk.Checkbutton(self, text="Enable", variable=self.vis_on)
        self.visualiser_checkbox.pack()

        self.visualiser_minsat = ttk.Scale(self, from_=0, to=65535, command=self.onScale_minsat)
        self.visualiser_minsat.pack()

        self.visualiser_minsat_label = ttk.Label(self, text=0, textvariable=self.minsat)
        self.visualiser_minsat_label.pack()

        self.visualiser_maxsat = ttk.Scale(self, from_=0, to=65535, command=self.onScale_maxsat)
        self.visualiser_maxsat.pack()

        self.visualiser_maxsat_label = ttk.Label(self, text=0, textvariable=self.maxsat)
        self.visualiser_maxsat_label.pack()

    def onScale_minsat(self, val):

        v = int(float(val))
        self.minsat.set(v)

    def onScale_maxsat(self, val):

        v = int(float(val))
        self.maxsat.set(v)
